<?php
require_once(dirname(__FILE__).'/MySQLBase.php');

//  TODO
////////////
//  FindAllByRelation :: prendre en compte primaryKey
//  General :: faire un outil de requte
//  Requete :: pouvoir rcup les donnes de l'objet ET des relations
//  Save :: compar avec les donnes initial pour UPDATER que les donnes modifis
//  options :: pouvoir automatiser la gestion des champs date_create/date_update
//  options :: pouvoir dsactiver la verification de la configuration des models
//  Generator :: effectuer des tests (duplication de class, variable, relations, ..., reference  une class, variable qui n'existe pas, verifier qu'au moins une clef primaire est dfinis)
//  Generator :: pouvoir dfinir le chemin de destination des models gnrs
//  Error :: externaliser l'outil d'affichage d'erreur, pouvoir l'utiliser dans un contexte static

abstract class MySQLObject
{
  private $mysql;
  private $table;
  private $primaryKey=array();
  private $columns=array();
  private $property=array();
  private $relations=array();
  private $options=array();


  public function __construct()
  {
    $this->mysql = MySQLBase::GetInstance();
    $this->Configure();
    $this->CheckConfigure();
  }

  public function __set($name, $value)
  {
    if(array_key_exists($name, $this->columns))
      $this->SetProperty($name, $value);
    elseif(array_key_exists($name, $this->relations))
      $this->SetRelation($name, $value);
    else
      $this->Error('"'.$name.'" n\'est pas dfinis dans le modle');
  }

  public function __get($name)
  {
    if(array_key_exists($name, $this->property))
      return $this->GetProperty($name);
    elseif(array_key_exists($name, $this->relations))
      return $this->GetRelation($name);
    else
      $this->Error('"'.$name.'" n\'existe pas');
  }


  protected abstract function Configure();

  private function CheckConfigure()
  {
    if(!count($this->primaryKey))
      $this->Error('Impossible de continuer, aucune primaryKey de dfinis');
  }

  protected function SetTable($name)
  {
    $this->table = $name;
  }

  protected function SetColumn($name, $params)
  {
    if($this->NameExist($name))
      $this->Error('"'.$name.'" est dj utilis comme nom de variable ou de relation');

    if(isset($params['primary']) && $params['primary'])
      $this->primaryKey[] = $name;

    $this->columns[$name] = $params;
  }

  protected function SetupRelation($name, $params)
  {
    if($this->NameExist($name))
      $this->Error('"'.$name.'" est dj utilis comme nom de variable ou de relation');

    switch($params['type'])
    {
      case 'oneToOne':
        $this->relations[$name] = array('params'=>$params, 'data'=>null);
        break;
      case 'oneToMany':
        $this->relations[$name] = array('params'=>$params, 'data'=>array());
        break;
      case 'manyToMany':
        $this->relations[$name] = array('params'=>$params, 'data'=>array());
        break;
      default:
        $this->Error('Type de relation inconnus');
        break;
    }
  }

  private function SetProperty($name, $value)
  {
    $this->property[$name] = $value;
  }

  private function SetRelation($name, $value)
  {
  }

  private function GetProperty($name)
  {
    return $this->property[$name];
  }

  private function GetRelation($name)
  {
    $relation = &$this->relations[$name];

    if((!is_array($relation['data']) && $relation['data'] !== null) || (is_array($relation['data']) && count($relation['data'])))
      return $relation['data'];

    switch($relation['params']['type'])
    {
      case 'oneToOne':
        return $relation['data'] = $this->FindOneBy($relation['params']['class'], $relation['params']['foreign'], $this->$relation['params']['local']);
        break;
      case 'oneToMany':
        return $relation['data'] = $this->FindAll($relation['params']['class'], $relation['params']['foreign'], $this->$relation['params']['local']);
        break;
      case 'manyToMany':
        return $relation['data'] = $this->FindAllByRelation($relation);
        break;
    }
  }

  private function IsPrimaryKeyDefine()
  {
    foreach($this->primaryKey as $key)
      if(!isset($this->property[$key]))
        return false;

    return true;
  }


  public static function FindOneBy($class, $field, $value)
  {
    $tmpObject = new $class;

    $sql = 'SELECT '.implode(', ', array_keys($tmpObject->columns)).
    ' FROM '.$tmpObject->table.
    ' WHERE '.$field.' = "'.mysql_real_escape_string($value).'"';
    $tmpObject->mysql->SetRequest($sql);

    return self::SetObjectFromQuery($class, $tmpObject->mysql->Fetch());
  }

  public static function FindAll($class, $field='', $value='')
  {
    $tmpObject = new $class;
    $tmpArray = array();

    $sql = 'SELECT '.implode(', ', array_keys($tmpObject->columns)).
    ' FROM '.$tmpObject->table;
    if($field && $value)
      $sql .= ' WHERE '.$field.' = "'.mysql_real_escape_string($value).'"';
    $sql .= ' ORDER BY id ASC';
    $tmpObject->mysql->SetRequest($sql);

    return self::SetObjectFromQuery($class, $tmpObject->mysql->FetchAll());
  }

  private function FindAllByRelation($relation)
  {
    $refObject = new $relation['params']['refClass'];
    $tmpObject = new $relation['params']['class'];

    $sql = 'SELECT t3.'.implode(', t3.', array_keys($tmpObject->columns)).
    ' FROM '.$this->table.' t1, '.$refObject->table.' t2, '.$tmpObject->table.' t3'.
    ' WHERE t1.id = '.$this->id.
    ' AND t1.id = t2.'.$relation['params']['local'].
    ' AND t2.'.$relation['params']['foreign'].' = t3.id'.
    ' ORDER BY t3.id ASC';
    $tmpObject->mysql->SetRequest($sql);

    return $this->SetObjectFromQuery($relation['params']['class'], $tmpObject->mysql->FetchAll());
  }

  protected static function SetObjectFromQuery($class, $values)
  {
    if(($values !== false && !count($values)) || (isset($values[0]) && is_array($values[0])))
    {
      $tmpArray = array();
      foreach($values as $tmpResult)
      {
        $tmp = new $class;
        foreach($tmpResult as $key=>$value)
          $tmp->property[$key] = $value;

        if(!$tmp->IsPrimaryKeyDefine())
          $tmp->Error('Impossible de continuer, les primaryKey doivent tre dfinis');
        $tmpArray[] = $tmp;
      }
      return $tmpArray;
    }
    else
    {
      if(!$values)
        return null;

      $tmpObject = new $class;
      foreach($values as $key=>$value)
        $tmpObject->property[$key] = $value;

      if(!$tmpObject->IsPrimaryKeyDefine())
        $tmpObject->Error('Impossible de continuer, les primaryKey doivent tre dfinis');
      return $tmpObject;
    }
  }


  public function Save()
  {
    $this->CleanProperty();

    if(!$this->IsPrimaryKeyDefine())
    {
      $sql = 'INSERT INTO '.$this->table.' ('.implode(', ',array_keys($this->property)).')'.
      ' VALUES ("'.implode('", "',array_values($this->property)).'")';
    }
    else
    {
      $sql = 'UPDATE '.$this->table.' SET ';
      foreach($this->property as $key=>$value)
      {
        if(isset($notFirst))
          $sql .= ', ';
        $sql .= $key.' = "'.$value.'"';
        $notFirst = true;
      }

      $notFirst = false;
      $sql .= ' WHERE ';
      foreach($this->primaryKey as $key)
      {
        if($notFirst)
          $sql .= ' AND ';
        $sql .= $key.' = "'.$this->property[$key].'"';
        $notFirst = true;
      }
    }
    $this->mysql->SetRequest($sql);

    if(isset($this->columns['id'], $this->columns['id']['autoincrement']) && !isset($this->property['id']))
      $this->property['id'] = $this->mysql->LastInsertedId();
  }

  public function Link($relationName, $objectToLink)
  {
    if(!array_key_exists($relationName, $this->relations))
      $this->Error('La relation "'.$relationName.'" n\'existe pas');

    $this->$relationName;
    $relation = &$this->relations[$relationName];

    switch($relation['params']['type'])
    {
      case 'oneToOne':
        $this->$relation['params']['local'] = $objectToLink->$relation['params']['foreign'];
        $this->Save();
        $relation['data'] = $objectToLink;
        break;
      case 'oneToMany':
        $objectToLink->$relation['params']['foreign'] = $this->$relation['params']['local'];
        $objectToLink->Save();
        $relation['data'][] = $objectToLink;
        break;
      case 'manyToMany':
        $refClass = new $relation['params']['refClass'];
        $refClass->$relation['params']['local'] = $this->id;
        $refClass->$relation['params']['foreign'] = $objectToLink->id;
        $refClass->Save();
        $relation['data'][] = $objectToLink;
        break;
    }
  }


  public function Delete()
  {
    if(!$this->IsPrimaryKeyDefine())
      $this->Error('Impossible de supprimer l\'objet');

    $this->DeleteRelation();

    $sql = 'DELETE FROM '.$this->table.' WHERE id = '.$this->id;
    $this->mysql->SetRequest($sql);

    return $this->mysql->CountModified();
  }

  public function DeleteAllBy($field, $value)
  {
    if(!array_key_exists($field, $this->columns))
      $this->Error('Le champs '.$field.' n\'est pas dclar dans le modle');

    $sql = 'DELETE FROM '.$this->table.' WHERE '.$field.' = "'.mysql_real_escape_string($value).'"';
    $this->mysql->SetRequest($sql);

    return $this->mysql->CountModified();
  }

  private function DeleteRelation()
  {
    if(!count($this->relations))
      return;

    foreach($this->relations as $name=>$relation)
    {
      $methodName = 'DeleteRelation'.ucfirst($relation['params']['type']);
      if(isset($relation['params']['cascadeDelete']) && $relation['params']['cascadeDelete'] && method_exists($this, $methodName))
        $this->$methodName($name);
    }
  }

  private function DeleteRelationOneToMany($relationName)
  {
    foreach($this->$relationName as $relation)
      $relation->Delete();

    unset($this->relations[$relationName]);
  }

  private function DeleteRelationManyToMany($relationName)
  {
    $relation = &$this->relations[$relationName];

    $tmpObject = new $relation['params']['refClass'];
    $tmpObject->DeleteAllBy($relation['params']['local'], $this->id);
    unset($this->relations[$relationName]);
  }


  private function CleanProperty()
  {
    foreach($this->property as $key=>$val)
      $this->property[$key] = mysql_real_escape_string($val);
  }

  private function NameExist($name)
  {
    if(array_key_exists($name, $this->columns) || array_key_exists($name, $this->property))
      return true;
    else
      return false;
  }

  private function Error($error='')
  {
    $listBackTrace = debug_backtrace();
    array_shift($listBackTrace);
    $listBackTrace = array_reverse($listBackTrace);

    $html_backTrace = '';
    foreach($listBackTrace as $backTrace)
    {
      $html_backTrace .= '
        <li>
          <strong>file</strong> : "'.$backTrace['file'].'" -
          <strong>line</strong> : '.$backTrace['line'].' -
          <strong>function</strong> : '.$backTrace['function'].'
        </li>
      ';
    }

    $html = '
      <h1>Erreur</h1>
      <ul>
        <li><strong>Object</strong> : '.get_class($this).'</li>
        <li><strong>Erreur</strong> : '.$error.'</li>
        <li><strong>BackTrace</strong> :
          <ul>
            '.$html_backTrace.'
          </ul>
        </li>
      </ul>
    ';
    exit($html);
  }


  public static function LoadModels($path=null)
  {
    if($path === null)
      $path = dirname(__FILE__).'/../models';

    if(in_array('generated', scandir($path)))
      self::LoadModels($path.'/generated');

    $directory = opendir($path) or die();
    while($entry = @readdir($directory))
    {
      if(!is_dir($path.'/'.$entry) && $entry != '.' && $entry != '..')
        include_once($path.'/'.$entry);
      /*elseif(is_dir($path.'/'.$entry) && $entry != '.' && $entry != '..')
        self::LoadModels($path.'/'.$entry);*/
    }
    closedir($directory);
  }
}

?>
